# date4/date.rb: Written by Tadayoshi Funaba 1998-2006
# $Id: date.rb,v 1.14 2006-12-30 21:52:50+09 tadf Exp $

require 'rational'
require 'date4/infinity'
require 'date4/const'
require 'date4/base'
require 'date4/misc'
require 'date4/delta'

class Date4

  include Comparable

  extend  Base
  include Base

  def self.julian_leap? (y) y % 4 == 0 end
  def self.gregorian_leap? (y) y % 4 == 0 && y % 100 != 0 || y % 400 == 0 end

  class << self; alias_method :leap?, :gregorian_leap? end
  class << self; alias_method :new!, :new end

  def self.valid_jd? (*args)
    jd, sg, = parsearg([[nil, :jd],
			[ITALY, :start]], args)
    !!_valid_jd?(jd, sg)
  end

  def self.valid_ordinal? (*args)
    y, d, sg, = parsearg([[nil, :y, :year],
			  [nil, :d, :day, :yday],
			  [ITALY, :start]], args)
    !!_valid_ordinal?(y, d, sg)
  end

  def self.valid_civil? (*args)
    y, m, d, sg, = parsearg([[nil, :y, :year],
			     [nil, :m, :mon, :month],
			     [nil, :d, :day, :mday],
			     [ITALY, :start]], args)
    !!_valid_civil?(y, m, d, sg)
  end

  class << self; alias_method :valid_date?, :valid_civil? end

  def self.valid_commercial? (*args)
    y, w, d, sg, = parsearg([[nil, :y, :year, :cwyear],
			     [nil, :w, :week, :cweek],
			     [nil, :d, :day, :cwday],
			     [ITALY, :start]], args)
    !!_valid_commercial?(y, w, d, sg)
  end

  def self.valid_weeknum? (*args)
    y, w, d, f, sg, = parsearg([[nil, :y, :year],
				[nil, :w, :week],
				[nil, :d, :day, :wday],
				[nil, :f, :first],
				[ITALY, :start]], args)
    !!_valid_commercial?(y, w, d, k, sg)
  end

  private_class_method :valid_weeknum?

  def self.valid_nth_kday? (*args)
    y, m, n, k, sg, = parsearg([[nil, :y, :year],
				[nil, :m, :mon, :month],
				[nil, :n, :nth],
				[nil, :k, :kday],
				[ITALY, :start]], args)
    !!_valid_nth_kday?(y, m, n, k, sg)
  end

#  private_class_method :valid_nth_kday?

  def self.valid_time? (*args)
    h, min, s, = parsearg([[nil, :h, :hour],
			   [nil, :min, :minute],
			   [nil, :s, :sec, :second]], args)

    !!_valid_time?(h, min, s)
  end

  private_class_method :valid_time?

  def self.jd(*args)
    jd, sg, = parsearg([[0, :jd],
			[ITALY, :start]], args)
    jd = _valid_jd?(jd, sg)
    new!([jd, 0], 0, sg)
  end

  def self.ordinal(*args)
    y, d, sg, = parsearg([[-4712, :y, :year],
			  [1, :d, :day, :yday],
			  [ITALY, :start]], args)

    unless jd = _valid_ordinal?(y, d, sg)
      raise ArgumentError, 'invalid date'
    end
    new!([jd, 0], 0, sg)
  end

  def self.civil(*args)
    y, m, d, sg, = parsearg([[-4712, :y, :year],
			     [1, :m, :mon, :month],
			     [1, :d, :day, :mday],
			     [ITALY, :start]], args)

    unless jd = _valid_civil?(y, m, d, sg)
      raise ArgumentError, 'invalid date'
    end
    new!([jd, 0], 0, sg)
  end

  class << self; alias_method :new, :civil end

  def self.commercial(*args)
    y, w, d, sg, = parsearg([[-4712, :y, :year, :cwyear],
			     [1, :w, :week, :cweek],
			     [1, :d, :day, :cwday],
			     [ITALY, :start]], args)

    unless jd = _valid_commercial?(y, w, d, sg)
      raise ArgumentError, 'invalid date'
    end
    new!([jd, 0], 0, sg)
  end

  def self.weeknum(*args)
    y, w, d, f, sg, = parsearg([[-4712, :y, :year],
				[0, :w, :week],
				[1, :d, :day, :wday],
				[0, :f, :first],
				[ITALY, :start]], args)

    unless jd = _valid_weeknum?(y, w, d, f, sg)
      raise ArgumentError, 'invalid date'
    end
    new!(jd_to_ajd(jd, 0, 0), 0, sg)
  end

  private_class_method :weeknum

  def self.nth_kday(*args)
    y, m, n, k, sg, = parsearg([[-4712, :y, :year],
				[1, :m, :mon, :month],
				[1, :n, :nth],
				[1, :k, :kday, :wday],
				[ITALY, :start]], args)
    unless jd = _valid_nth_kday?(y, m, n, k, sg)
      raise ArgumentError, 'invalid date'
    end
    new!([jd, 0], 0, sg)
  end

#  private_class_method :nth_kday

  def self.rewrite_frags(elem)
    elem ||= {}
    if seconds = elem[:seconds]
      d,   fr = seconds.divmod(86400)
      h,   fr = fr.divmod(3600)
      min, fr = fr.divmod(60)
      s,   fr = fr.divmod(1)
      elem[:jd] = UNIXEPOCH + d
      elem[:hour] = h
      elem[:min] = min
      elem[:sec] = s
      elem[:sec_fraction] = fr
      elem.delete(:seconds)
      elem.delete(:offset)
    end
    elem
  end

  private_class_method :rewrite_frags

  def self.complete_frags(elem)
    i = 0
    g = [[:time, [:hour, :min, :sec]],
	 [nil, [:jd]],
	 [:ordinal, [:year, :yday, :hour, :min, :sec]],
	 [:civil, [:year, :mon, :mday, :hour, :min, :sec]],
	 [:commercial, [:cwyear, :cweek, :cwday, :hour, :min, :sec]],
	 [:wday, [:wday, :hour, :min, :sec]],
	 [:wnum0, [:year, :wnum0, :wday, :hour, :min, :sec]],
	 [:wnum1, [:year, :wnum1, :wday, :hour, :min, :sec]],
	 [nil, [:cwyear, :cweek, :wday, :hour, :min, :sec]],
	 [nil, [:year, :wnum0, :cwday, :hour, :min, :sec]],
	 [nil, [:year, :wnum1, :cwday, :hour, :min, :sec]]].
      collect{|k, a| e = elem.values_at(*a).compact; [k, a, e]}.
      select{|k, a, e| e.size > 0}.
      sort_by{|k, a, e| [e.size, i -= 1]}.last

    d = nil

    if g && g[0] && (g[1].size - g[2].size) != 0
      d ||= Date4.today

      case g[0]
      when :ordinal
	elem[:year] ||= d.year
	elem[:yday] ||= 1
      when :civil
	g[1].each do |e|
	  break if elem[e]
	  elem[e] = d.__send__(e)
	end
	elem[:mon]  ||= 1
	elem[:mday] ||= 1
      when :commercial
	g[1].each do |e|
	  break if elem[e]
	  elem[e] = d.__send__(e)
	end
	elem[:cweek] ||= 1
	elem[:cwday] ||= 1
      when :wday
	elem[:jd] ||= (d - d.wday + elem[:wday]).jd
      when :wnum0
	g[1].each do |e|
	  break if elem[e]
	  elem[e] = d.__send__(e)
	end
	elem[:wnum0] ||= 0
	elem[:wday]  ||= 0
      when :wnum1
	g[1].each do |e|
	  break if elem[e]
	  elem[e] = d.__send__(e)
	end
	elem[:wnum1] ||= 0
	elem[:wday]  ||= 0
      end
    end

    if g && g[0] == :time
      if self <= DateTime4
	d ||= Date4.today
	elem[:jd] ||= d.jd
      end
    end

    elem[:hour] ||= 0
    elem[:min]  ||= 0
    elem[:sec]  ||= 0
    elem[:sec] = [elem[:sec], 59].min

    elem
  end

  private_class_method :complete_frags

  def self.valid_date_frags?(elem, sg)
    catch :jd do
      a = elem.values_at(:jd)
      if a.all?
	if jd = _valid_jd?(*(a << sg))
	  throw :jd, jd
	end
      end

      a = elem.values_at(:year, :yday)
      if a.all?
	if jd = _valid_ordinal?(*(a << sg))
	  throw :jd, jd
	end
      end

      a = elem.values_at(:year, :mon, :mday)
      if a.all?
	if jd = _valid_civil?(*(a << sg))
	  throw :jd, jd
	end
      end

      a = elem.values_at(:cwyear, :cweek, :cwday)
      if a[2].nil? && elem[:wday]
	a[2] = elem[:wday].nonzero? || 7
      end
      if a.all?
	if jd = _valid_commercial?(*(a << sg))
	  throw :jd, jd
	end
      end

      a = elem.values_at(:year, :wnum0, :wday)
      if a[2].nil? && elem[:cwday]
	a[2] = elem[:cwday] % 7
      end
      if a.all?
	if jd = _valid_weeknum?(*(a << 0 << sg))
	  throw :jd, jd
	end
      end

      a = elem.values_at(:year, :wnum1, :wday)
      if a[2]
	a[2] = (a[2] - 1) % 7
      end
      if a[2].nil? && elem[:cwday]
	a[2] = (elem[:cwday] - 1) % 7
      end
      if a.all?
	if jd = _valid_weeknum?(*(a << 1 << sg))
	  throw :jd, jd
	end
      end
    end
  end

  private_class_method :valid_date_frags?

  def self.valid_time_frags? (elem)
    h, min, s = elem.values_at(:hour, :min, :sec)
    _valid_time?(h, min, s)
  end

  private_class_method :valid_time_frags?

  def self.new_by_frags(elem, sg)
    elem = rewrite_frags(elem)
    elem = complete_frags(elem)
    unless jd = valid_date_frags?(elem, sg)
      raise ArgumentError, 'invalid date'
    end
    new!([jd, 0], 0, sg)
  end

  private_class_method :new_by_frags

  def self.strptime(*args)
    str, fmt, sg, lc, = parsearg([['-4712-01-01', :d, :date],
				  ['%F', :fmt, :format],
				  [ITALY, :start],
				  [Format::Locale::C, :lc, :locale]], args)
    elem = _strptime(str, fmt, lc)
    new_by_frags(elem, sg)
  end

  def self.parse(*args)
    str, comp, sg, lc, = parsearg([['-4712-01-01', :d, :date],
				   [false, :comp, :complete],
				   [ITALY, :start],
				   [Format::Locale::C, :lc, :locale]], args)
    elem = _parse(str, comp, lc)
    new_by_frags(elem, sg)
  end

  def initialize(ajd=0, tz=0, sg=ITALY)
    @sg = sg
    case ajd
    when Numeric
      case tz
      when Numeric; @of = tz;			      @tz = nil
      else;	    @of = tz.lookup(ajd, false)[:of]; @tz = tz
      end
      @ajd = ajd
    when Array
      jd, fr, = ajd
      fajd = jd_to_ajd(jd, fr, 0)
      case tz
      when Numeric; @of = tz;			      @tz = nil
      else;	    @of = tz.lookup(fajd, true)[:of]; @tz = tz
      end
      @ajd = jd_to_ajd(jd, fr, @of)
      case tz
      when Numeric
      else
	unless @of == tz.lookup(@ajd, false)[:of]
	  raise ArgumentError, 'invalid date'
	end
      end
    end
  end

  def ajd() @ajd end
  def amjd() ajd_to_amjd(@ajd) end

  once :amjd

  def jd() ajd_to_jd(@ajd, @of)[0] end
  def day_fraction() ajd_to_jd(@ajd, @of)[1] end
  def mjd() jd_to_mjd(jd) end
  def ld() jd_to_ld(jd) end

  once :jd, :day_fraction, :mjd, :ld

  def civil() jd_to_civil(jd, @sg) end
  def ordinal() jd_to_ordinal(jd, @sg) end
  def commercial() jd_to_commercial(jd, @sg) end
  def weeknum0() jd_to_weeknum(jd, 0, @sg) end
  def weeknum1() jd_to_weeknum(jd, 1, @sg) end

  once :civil, :ordinal, :commercial, :weeknum0, :weeknum1
  private :civil, :ordinal, :commercial, :weeknum0, :weeknum1

  def year() civil[0] end
  def yday() ordinal[1] end
  def mon() civil[1] end
  def mday() civil[2] end

  alias_method :month, :mon
  alias_method :day, :mday

  def wnum0() weeknum0[1] end
  def wnum1() weeknum1[1] end

  private :wnum0, :wnum1

  def time() day_fraction_to_time(day_fraction) end

  once :time
  private :time

  def hour() time[0] end
  def min() time[1] end
  def sec() time[2] end
  def sec_fraction() time[3] end

  private :hour, :min, :sec, :sec_fraction

  def tzlookup() @tz.lookup(@ajd, false) end

  once :tzlookup
  private :tzlookup

  def zone() (@tz && tzlookup[:zone]) || strftime('%:z') end

  once :zone
  private :zone

  def cwyear() commercial[0] end
  def cweek() commercial[1] end
  def cwday() commercial[2] end

  def wday() jd_to_wday(jd) end

  once :wday

  def utc? () @tz && tzlookup[:utc] end
  def dst? () @tz && tzlookup[:dst] end

  once :utc?, :dst?

=begin
  MONTHNAMES.each_with_index do |n, i|
    if n
      define_method(n.downcase + '?'){mon == i}
    end
  end
=end

  DAYNAMES.each_with_index do |n, i|
    define_method(n.downcase + '?'){wday == i}
  end

  def nth_kday? (*args)
    n, k, = parsearg([[nil, :n, :nth],
		      [nil, :k, :kday]], args)
    k == wday && jd == nth_kday_to_jd(year, mon, n, k, start)
  end

#  private :nth_kday?

  def julian? () jd < @sg end
  def gregorian? () !julian? end

  once :julian?, :gregorian?

  def fix_style
    if julian?
    then self.class::JULIAN
    else self.class::GREGORIAN end
  end

  private :fix_style

  def leap?
    jd_to_civil(civil_to_jd(year, 3, 1, fix_style) - 1,
		fix_style)[-1] == 29
  end

  once :leap?

  def start() @sg end

  def new_start(sg=self.class::ITALY)
    self.class.new!(@ajd, @tz || @of, sg)
  end

  def italy() new_start(self.class::ITALY) end
  def england() new_start(self.class::ENGLAND) end
  def julian() new_start(self.class::JULIAN) end
  def gregorian() new_start(self.class::GREGORIAN) end

  def offset() @of end

  alias_method :timezone, :offset
  private      :timezone, :offset

  def new_offset(of=0)
    if String === of
      of = (zone_to_diff(of) || 0).to_r/86400
    end
    self.class.new!(@ajd, of, @sg)
  end

  alias_method :new_timezone, :new_offset
  private      :new_timezone, :new_offset

  def + (n)
    case n
    when Numeric; return self.class.new!(@ajd + n.to_r, @tz || @of, @sg)
    else
      l, r = n.coerce(self)
      return l + r
    end
  end

  def - (x)
    case x
    when Numeric; return self.class.new!(@ajd - x.to_r, @tz || @of, @sg)
#    when Date4;   return @ajd - x.ajd
    when Date4;   return Delta.diff(self, x)
    else
      l, r = x.coerce(self)
      return l - r
    end
  end

  def <=> (other)
    case other
    when Numeric; return @ajd <=> other
    when Date4;   return @ajd <=> other.ajd
    else
      begin
	l, r = other.coerce(self)
	return l <=> r
      rescue NoMethodError
      end
    end
    nil
  end

  def === (other)
    case other
    when Numeric; return jd == other
    when Date4;   return jd == other.jd
    else
      l, r = other.coerce(self)
      return l === r
    end
    false
  end

  def next_day(n=1) self + n end
  def prev_day(n=1) self - n end

  def next() next_day end

  alias_method :succ, :next

  def >> (n)
    y, m = (year * 12 + (mon - 1) + n).divmod(12)
    m,   = (m + 1).divmod(1)
    d = mday
    d -= 1 until jd2 = _valid_civil?(y, m, d, @sg)
    self + (jd2 - jd)
  end

  def << (n) self >> -n end

  def next_month(n=1) self >> n end
  def prev_month(n=1) self << n end

  def next_year(n=1) self >> n * 12 end
  def prev_year(n=1) self << n * 12 end

  require 'enumerator'

  def step(limit, step=1)
    unless block_given?
      return to_enum(:step, limit, step)
    end
    da = self
    op = %w(- <= >=)[step <=> 0]
    while da.__send__(op, limit)
      yield da
      da += step
    end
    self
  end

  def upto  (max, &block) step(max, +1, &block) end
  def downto(min, &block) step(min, -1, &block) end

  def eql? (other) Date4 === other && self == other end
  def hash() @ajd.hash end

  def inspect
    of = @of.to_s
    of += format('(%s)', zone) if @tz
    format('#<%s: %s,%s,%s>', self.class, @ajd, of, @sg)
  end

  def to_s() strftime('%F') end

  def marshal_dump() [@ajd, @tz || @of, @sg] end
  def marshal_load(a) initialize(*a) end

end

class Date4

  remove_const :Base
  remove_const :Once
  remove_const :ParseArg

end
