# date4/delta.rb: Written by Tadayoshi Funaba 2004-2006
# $Id: delta.rb,v 1.4 2006-12-30 21:49:56+09 tadf Exp $

require 'rational'
require 'date4/delta/const'
require 'date4/delta/parser'
require 'date4/delta/misc'

class Date4

  class Delta

    include Comparable

    def self.delta_to_dhms(delta)
      sign = if delta < 0 then -1 else +1 end
      fr = delta.abs
      d,   fr = fr.divmod(1.to_r)
      h,   fr = fr.divmod(1.to_r/24)
      min, fr = fr.divmod(1.to_r/1440)
      s,   fr = fr.divmod(1.to_r/86400)
      return [d, h, min, s, fr].collect{|x| sign * x}
    end

    def self.dhms_to_delta(d, h, min, s, fr)
      d + h.to_r/24 + min.to_r/1440 + (s + fr).to_r/86400
    end

    def initialize(*args)
      d, h, min, s = parsearg([[0, :d, :days],
			       [0, :h, :hours],
			       [0, :min, :mins, :minutes],
			       [0, :s, :secs, :seconds]], args)
      @in_days = self.class.dhms_to_delta(d, h, min, s, 0)
    end

    UNITS.each_key do |k|
      module_eval <<-"end;"
	def self.#{k}s(n=1)
	  new(:d=>n * UNITS['#{k}'])
	end
      end;
    end

    class << self; alias_method :mins, :minutes end
    class << self; alias_method :secs, :seconds end

    def self.parse(*args)
      str, = parsearg([[nil, :d, :delta]], args)
      d = begin (@@pa ||= Parser.new).parse(str)
	  rescue Racc::ParseError
	    raise ArgumentError, 'syntax error'
	  end
      new(:d=>d)
    end

    def self.diff(d1, d2) new(d1.ajd - d2.ajd) end

    def dhms() self.class.delta_to_dhms(in_days) end

    once :dhms

    def days() dhms[0] end
    def hours() dhms[1] end
    def minutes() dhms[2] end
    def seconds() dhms[3] end
    def second_fractions() dhms[4] end

    alias_method :mins, :minutes
    alias_method :secs, :seconds
    alias_method :sec_fractions, :second_fractions

    UNITS.each_key do |k|
      module_eval <<-"end;"
	def in_#{k}s(u=1)
	  @in_days / (u * UNITS['#{k}'])
	end
      end;
    end

    alias_method :in_mins, :in_minutes
    alias_method :in_secs, :in_seconds

    def abs() self.class.new(in_days.abs) end

    def zero?() in_days.zero? end
    def nonzero?() unless zero? then self end end

    def integer? () in_days.integer? end

    def -@ () self.class.new(-in_days) end
    def +@ () self.class.new(+in_days) end

    def dxnd (m, n)
      case n
      when Numeric; return self.class.new(in_days.__send__(m, n))
      when Delta; return self.class.new(in_days.__send__(m, n.in_days))
      else
	l, r = n.coerce(self)
	return l.__send__(m, r)
      end
    end

    private :dxnd

    def + (n) dxnd(:+, n) end
    def - (n) dxnd(:-, n) end

    def dxn (m, n)
      case n
      when Numeric; return self.class.new(in_days.__send__(m, n))
      else
	l, r = n.coerce(self)
	return l.__send__(m, r)
      end
    end

    private :dxn

    def * (n) dxn(:*, n) end
    def / (n) dxn(:/, n) end
    def % (n) dxn(:%, n) end

    def div(n) dxn(:div, n) end
    def modulo(n) dxn(:modulo, n) end
    def divmod(n) [div(n), modulo(n)] end

    def quotient(n)
      case n
      when Numeric; return self.class.new((in_days / n).truncate)
      else
	l, r = n.coerce(self)
	return l.__send__(m, r)
      end
    end

    def remainder(n) dxn(:remainder, n) end
    def quotrem(n) [quotient(n), remainder(n)] end

    def ** (n) dxn(:**, n) end
    def quo(n) dxn(:quo, n) end

    def <=> (other)
      case other
      when Numeric; return in_days <=> other
      when Delta;   return in_days <=> other.in_days
      else
	begin
	  l, r = other.coerce(self)
	  return l <=> r
	rescue NoMethodError
	end
      end
      nil
    end

    def coerce(other)
      case other
      when Date4; return other, in_days
      else
	super
      end
    end

    def eql? (other) Delta === other && self == other end
    def hash() in_days.hash end

    def ceil() in_days.ceil end
    def floor() in_days.floor end
    def round() in_days.round end
    def truncate() in_days.truncate end

    def to_i() in_days.to_i end
    def to_f() in_days.to_f end
    def to_r() in_days      end

    alias_method :to_int, :to_i

    def inspect() format('#<%s: %s>', self.class, in_days) end

    def to_s
      if in_days < 0 then '-' else  '+' end +
	format('P%dD%dH%dM%dS',
	       *([days, hours, mins, secs].collect{|x| x.abs}))
    end

  end

end
